==> intel-oneapi-base-toolkit-2025.2.0-3-x86_64.pkg.tar.zst: sycl/sparse_blas/source/csr_omatcopy.cpp <==
/*******************************************************************************
* Copyright (C) 2022 Intel Corporation
*
* This software and the related documents are Intel copyrighted  materials,  and
* your use of  them is  governed by the  express license  under which  they were
* provided to you (License).  Unless the License provides otherwise, you may not
* use, modify, copy, publish, distribute,  disclose or transmit this software or
* the related documents without Intel's prior written permission.
*
* This software and the related documents  are provided as  is,  with no express
* or implied  warranties,  other  than those  that are  expressly stated  in the
* License.
*******************************************************************************/

/*
*
*  Content:
*       This example demonstrates use of oneAPI Math Kernel Library (oneMKL)
*       DPCPP API, oneapi::mkl::sparse::omatcopy, to perform an out-of-place
*       copy/transpose operation on a sparse matrix and on a SYCL device (CPU,
*       GPU). This example uses sparse matrices in CSR format.
*
*           B = op(A)
*
*       where op() is defined by one of
*           oneapi::mkl::transpose::{nontrans,trans,conjtrans}
*
*       The supported floating point data types for omatcopy matrix data are:
*           float
*           double
*           std::complex<float>
*           std::complex<double>
*
*       The supported matrix formats for omatcopy are:
*           CSR
*           COO
*
*       Note that omatcopy does not support matrix format conversion (e.g. CSR
*       to COO), and both A and B must be in the same matrix format.
*
*******************************************************************************/

// stl includes
#include <algorithm>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <iterator>
#include <limits>
#include <list>
#include <vector>

#include "mkl.h"
#include "oneapi/mkl.hpp"
#include <sycl/sycl.hpp>

// local includes
#include "common_for_examples.hpp"
#include "./include/common_for_sparse_examples.hpp"

//
// Main example for Sparse Matrix copy/transpose: B = op(A)
//
template <typename dataType, typename intType>
int run_sparse_blas_example(sycl::queue &q)
{
    bool good = true;

    // handles for sparse matrix
    oneapi::mkl::sparse::matrix_handle_t csrA = nullptr;
    oneapi::mkl::sparse::matrix_handle_t csrB = nullptr;

    try {

        // Initialize data for Sparse Matrix
        oneapi::mkl::transpose opA      = oneapi::mkl::transpose::nontrans;
        oneapi::mkl::index_base a_index = oneapi::mkl::index_base::one;
        oneapi::mkl::index_base b_index = oneapi::mkl::index_base::zero;
        intType a_ind                   = a_index == oneapi::mkl::index_base::zero ? 0 : 1;
        intType b_ind                   = b_index == oneapi::mkl::index_base::zero ? 0 : 1;

        //
        // set up dimensions of matrix products
        //
        intType size = 4;

        intType a_nrows = size * size * size;
        intType a_ncols = a_nrows;
        intType a_nnz   = 27 * a_nrows;
        intType b_nrows = (opA == oneapi::mkl::transpose::nontrans ? a_nrows : a_ncols);
        intType b_ncols = (opA == oneapi::mkl::transpose::nontrans ? a_ncols : a_nrows);
        intType b_nnz   = a_nnz;


        //
        // setup A data locally in CSR format
        //
        std::vector<intType, mkl_allocator<intType, 64>> ia;
        std::vector<intType, mkl_allocator<intType, 64>> ja;
        std::vector<dataType, mkl_allocator<dataType, 64>> a;

        ia.resize(a_nrows + 1);
        ja.resize(27 * a_nrows);
        a.resize(27 * a_nrows);

        generate_sparse_matrix<dataType, intType>(size, ia, ja, a, a_ind);
        a_nnz = ia[a_nrows] - a_ind;


        //
        // setup B data locally in CSR format
        //
        std::vector<intType, mkl_allocator<intType, 64>> ib;
        std::vector<intType, mkl_allocator<intType, 64>> jb;
        std::vector<dataType, mkl_allocator<dataType, 64>> b;

        ib.resize(b_nrows + 1);
        jb.resize(27 * b_nrows);
        b.resize(27 * b_nrows);

        //
        // Execute Matrix Copy/Transpose
        //

        std::cout << "\n\t\tsparse::omatcopy parameters:\n";
        std::cout << "\t\t\topA = " << opA << std::endl;

        std::cout << "\t\t\tA_nrows = A_ncols = " << a_nrows << std::endl;
        std::cout << "\t\t\tB_nrows = B_ncols = " << b_nrows << std::endl;

        std::cout << "\t\t\tA_index = " << a_index << std::endl;
        std::cout << "\t\t\tB_index = " << b_index << std::endl;

        sycl::buffer<intType, 1> a_rowptr(ia.data(), ia.data() + a_nrows + 1);
        sycl::buffer<intType, 1> a_colind(ja.data(), ja.data() + a_nnz);
        sycl::buffer<dataType, 1> a_val(a.data(), a.data() +  a_nnz);
        sycl::buffer<intType, 1> b_rowptr(ib.data(), ib.data() + b_nrows + 1);
        sycl::buffer<intType, 1> b_colind(jb.data(), jb.data() + b_nnz);
        sycl::buffer<dataType, 1> b_val(b.data(), b.data() + b_nnz);

        oneapi::mkl::sparse::init_matrix_handle(&csrA);
        oneapi::mkl::sparse::init_matrix_handle(&csrB);

        oneapi::mkl::sparse::set_csr_data(q, csrA, a_nrows, a_ncols, a_index, a_rowptr, a_colind, a_val);
        oneapi::mkl::sparse::set_csr_data(q, csrB, b_nrows, b_ncols, b_index, b_rowptr, b_colind, b_val);

        // B = op(A)
        oneapi::mkl::sparse::omatcopy(q, opA, csrA, csrB);

        // clean up
        oneapi::mkl::sparse::release_matrix_handle(q, &csrA);
        oneapi::mkl::sparse::release_matrix_handle(q, &csrB);


        //
        // Post Processing
        //

        // Print portion of B solution
        {
            auto ib = b_rowptr.get_host_access(sycl::read_only);
            auto jb = b_colind.get_host_access(sycl::read_only);
            auto b  = b_val.get_host_access(sycl::read_only);
            std::cout << "B (= op(A)) matrix [first two rows]:" << std::endl;
            for (intType row = 0; row < std::min(static_cast<intType>(2), b_nrows); ++row) {
                for (intType j = ib[row] - b_ind; j < ib[row + 1] - b_ind; ++j) {
                    intType col = jb[j];
                    dataType val  = b[j];
                    std::cout << "B(" << row + b_ind << ", " << col << ") = " << val << std::endl;
                }
            }
        }

        good = true;

        q.wait_and_throw();
    }
    catch (sycl::exception const &e) {
        std::cout << "\t\tCaught synchronous SYCL exception:\n" << e.what() << std::endl;
        good = false;
    }
    catch (std::exception const &e) {
        std::cout << "\t\tCaught std exception:\n" << e.what() << std::endl;
        good = false;
    }

    q.wait();

    // backup cleaning of matrix handle and others for if exceptions happened
    if(csrA) oneapi::mkl::sparse::release_matrix_handle(q, &csrA);
    if(csrB) oneapi::mkl::sparse::release_matrix_handle(q, &csrB);

    q.wait();

    return good ? 0 : 1;
}

//
// Description of example setup, apis used and supported floating point type
// precisions
//
void print_example_banner()
{

    std::cout << "" << std::endl;
    std::cout << "###############################################################"
                 "#########"
              << std::endl;
    std::cout << "# Sparse Matrix Copy/Transpose/Conjugate-transpose Example: " << std::endl;
    std::cout << "# " << std::endl;
    std::cout << "#    B = op(A)" << std::endl;
    std::cout << "# " << std::endl;
    std::cout << "# where A and B are sparse matrices in CSR format" << std::endl;
    std::cout << "# " << std::endl;
    std::cout << "# Using apis:" << std::endl;
    std::cout << "#   sparse::omatcopy" << std::endl;
    std::cout << "# " << std::endl;
    std::cout << "#   sparse::init_matrix_handle" << std::endl;
    std::cout << "#   sparse::set_csr_data" << std::endl;
    std::cout << "#   sparse::release_matrix_handle" << std::endl;
    std::cout << "# " << std::endl;
    std::cout << "# Supported floating point type precisions:" << std::endl;
    std::cout << "#   float" << std::endl;
    std::cout << "#   double" << std::endl;
    std::cout << "#   std::complex<float>" << std::endl;
    std::cout << "#   std::complex<double>" << std::endl;
    std::cout << "# " << std::endl;
    std::cout << "###############################################################"
                 "#########"
              << std::endl;
    std::cout << std::endl;
}

//
// Main entry point for example.
//
// Dispatches to appropriate device types as set at build time with flag:
// -DSYCL_DEVICES_cpu -- only runs SYCL CPU implementation
// -DSYCL_DEVICES_gpu -- only runs SYCL GPU implementation
// -DSYCL_DEVICES_all (default) -- runs on all: cpu and gpu devices
//
//  For each device selected and each supported data type,
//  run_sparse_blas_example() is run with all supported data types,
//  if any fail, we move on to the next device.
//

int main(int argc, char **argv)
{
    print_example_banner();

    std::list<my_sycl_device_types> list_of_devices;
    set_list_of_devices(list_of_devices);

    int status = 0;
    for (auto it = list_of_devices.begin(); it != list_of_devices.end(); ++it) {
        try {
            sycl::device my_dev;
            bool my_dev_is_found = false;
            get_sycl_device(my_dev, my_dev_is_found, *it);

            if (my_dev_is_found) {
                std::cout << "Running tests on " << sycl_device_names[*it] << ".\n";

                // Catch asynchronous exceptions
                auto exception_handler = [](sycl::exception_list exceptions) {
                    for (std::exception_ptr const &e : exceptions) {
                        try {
                            std::rethrow_exception(e);
                        }
                        catch (sycl::exception const &e) {
                            std::cout << "Caught asynchronous SYCL exception: \n"
                                << e.what() << std::endl;
                        }
                    }
                };

                sycl::queue q(my_dev, exception_handler);

                std::cout << "\tRunning with single precision real data type:" << std::endl;
                status |= run_sparse_blas_example<float, std::int32_t>(q);

                if (my_dev.get_info<sycl::info::device::double_fp_config>().size() != 0) {
                    std::cout << "\tRunning with double precision real data type:" << std::endl;
                    status |= run_sparse_blas_example<double, std::int32_t>(q);
                }

                std::cout << "\tRunning with single precision complex data type:" << std::endl;
                status |= run_sparse_blas_example<std::complex<float>, std::int32_t>(q);

                if (my_dev.get_info<sycl::info::device::double_fp_config>().size() != 0) {
                    std::cout << "\tRunning with double precision complex data type:" << std::endl;
                    status |= run_sparse_blas_example<std::complex<double>, std::int32_t>(q);
                }

            }
            else {
#ifdef FAIL_ON_MISSING_DEVICES
                std::cout << "No " << sycl_device_names[*it]
                    << " devices found; Fail on missing devices "
                    "is enabled.\n";
                return 1;
#else
                std::cout << "No " << sycl_device_names[*it] << " devices found; skipping "
                    << sycl_device_names[*it] << " tests.\n";
#endif
            }
        }
        catch (sycl::exception const &e) {
            std::cout << "\t\tCaught SYCL exception at driver level: \n" << e.what() << std::endl;
            continue; // stop with device, but move on to other devices
        }
        catch (std::exception const &e) {
            std::cout << "\t\tCaught std exception at driver level: \n" << e.what() << std::endl;
            continue; // stop with device, but move on to other devices
        }

    } // for device

    mkl_free_buffers();
    return status;
}

==> intel-oneapi-base-toolkit-2025.3.0-1-x86_64.pkg.tar.zst: sycl/sparse_blas/source/csr_omatcopy.cpp <==
/*******************************************************************************
* Copyright (C) 2022 Intel Corporation
*
* This software and the related documents are Intel copyrighted  materials,  and
* your use of  them is  governed by the  express license  under which  they were
* provided to you (License).  Unless the License provides otherwise, you may not
* use, modify, copy, publish, distribute,  disclose or transmit this software or
* the related documents without Intel's prior written permission.
*
* This software and the related documents  are provided as  is,  with no express
* or implied  warranties,  other  than those  that are  expressly stated  in the
* License.
*******************************************************************************/

/*
*
*  Content:
*       This example demonstrates use of oneAPI Math Kernel Library (oneMKL)
*       DPCPP API, oneapi::mkl::sparse::omatcopy, to perform an out-of-place
*       copy/transpose operation on a sparse matrix and on a SYCL device (CPU,
*       GPU). This example uses sparse matrices in CSR format.
*
*           B = op(A)
*
*       where op() is defined by one of
*           oneapi::mkl::transpose::{nontrans,trans,conjtrans}
*
*       The supported floating point data types for omatcopy matrix data are:
*           float
*           double
*           std::complex<float>
*           std::complex<double>
*
*       The supported matrix formats for omatcopy are:
*           CSR
*           COO
*
*       Note that omatcopy does not support matrix format conversion (e.g. CSR
*       to COO), and both A and B must be in the same matrix format.
*
*******************************************************************************/

// stl includes
#include <algorithm>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <iterator>
#include <limits>
#include <list>
#include <vector>

#include "mkl.h"
#include "oneapi/mkl.hpp"
#include <sycl/sycl.hpp>

// local includes
#include "common_for_examples.hpp"
#include "./include/common_for_sparse_examples.hpp"

//
// Main example for Sparse Matrix copy/transpose: B = op(A)
//
template <typename dataType, typename intType>
int run_sparse_blas_example(sycl::queue &q)
{
    bool good = true;

    // handles for sparse matrix
    oneapi::mkl::sparse::matrix_handle_t csrA = nullptr;
    oneapi::mkl::sparse::matrix_handle_t csrB = nullptr;

    try {

        // Initialize data for Sparse Matrix
        oneapi::mkl::transpose opA      = oneapi::mkl::transpose::nontrans;
        oneapi::mkl::index_base a_index = oneapi::mkl::index_base::one;
        oneapi::mkl::index_base b_index = oneapi::mkl::index_base::zero;
        intType a_ind                   = a_index == oneapi::mkl::index_base::zero ? 0 : 1;
        intType b_ind                   = b_index == oneapi::mkl::index_base::zero ? 0 : 1;

        //
        // set up dimensions of matrix products
        //
        intType size = 4;

        const std::int64_t a_nrows = size * size * size;
        const std::int64_t a_ncols = a_nrows;
        std::int64_t       a_nnz   = 27 * a_nrows;
        const std::int64_t b_nrows = (opA == oneapi::mkl::transpose::nontrans ? a_nrows : a_ncols);
        const std::int64_t b_ncols = (opA == oneapi::mkl::transpose::nontrans ? a_ncols : a_nrows);
        std::int64_t       b_nnz   = a_nnz;


        //
        // setup A data locally in CSR format
        //
        std::vector<intType, mkl_allocator<intType, 64>> ia;
        std::vector<intType, mkl_allocator<intType, 64>> ja;
        std::vector<dataType, mkl_allocator<dataType, 64>> a;

        ia.resize(a_nrows + 1);
        ja.resize(27 * a_nrows);
        a.resize(27 * a_nrows);

        generate_sparse_matrix<dataType, intType>(size, ia, ja, a, a_ind);
        a_nnz = ia[a_nrows] - a_ind;
        b_nnz = a_nnz;

        //
        // setup B data locally in CSR format
        //
        std::vector<intType, mkl_allocator<intType, 64>> ib;
        std::vector<intType, mkl_allocator<intType, 64>> jb;
        std::vector<dataType, mkl_allocator<dataType, 64>> b;

        ib.resize(b_nrows + 1);
        jb.resize(27 * b_nrows);
        b.resize(27 * b_nrows);

        //
        // Execute Matrix Copy/Transpose
        //

        std::cout << "\n\t\tsparse::omatcopy parameters:\n";
        std::cout << "\t\t\topA = " << opA << std::endl;

        std::cout << "\t\t\tA_nrows = A_ncols = " << a_nrows << std::endl;
        std::cout << "\t\t\tB_nrows = B_ncols = " << b_nrows << std::endl;

        std::cout << "\t\t\tA_index = " << a_index << std::endl;
        std::cout << "\t\t\tB_index = " << b_index << std::endl;

        sycl::buffer<intType, 1> a_rowptr(ia.data(), ia.data() + a_nrows + 1);
        sycl::buffer<intType, 1> a_colind(ja.data(), ja.data() + a_nnz);
        sycl::buffer<dataType, 1> a_val(a.data(), a.data() +  a_nnz);
        sycl::buffer<intType, 1> b_rowptr(ib.data(), ib.data() + b_nrows + 1);
        sycl::buffer<intType, 1> b_colind(jb.data(), jb.data() + b_nnz);
        sycl::buffer<dataType, 1> b_val(b.data(), b.data() + b_nnz);

        oneapi::mkl::sparse::init_matrix_handle(&csrA);
        oneapi::mkl::sparse::init_matrix_handle(&csrB);

        oneapi::mkl::sparse::set_csr_data(q, csrA, a_nrows, a_ncols, a_nnz, a_index, a_rowptr, a_colind, a_val);
        oneapi::mkl::sparse::set_csr_data(q, csrB, b_nrows, b_ncols, b_nnz, b_index, b_rowptr, b_colind, b_val);

        // B = op(A)
        oneapi::mkl::sparse::omatcopy(q, opA, csrA, csrB);

        // clean up
        oneapi::mkl::sparse::release_matrix_handle(q, &csrA);
        oneapi::mkl::sparse::release_matrix_handle(q, &csrB);


        //
        // Post Processing
        //

        // Print portion of B solution
        {
            auto ib = b_rowptr.get_host_access(sycl::read_only);
            auto jb = b_colind.get_host_access(sycl::read_only);
            auto b  = b_val.get_host_access(sycl::read_only);
            std::cout << "B (= op(A)) matrix [first two rows]:" << std::endl;
            for (intType row = 0; row < std::min(static_cast<std::int64_t>(2), b_nrows); ++row) {
                for (intType j = ib[row] - b_ind; j < ib[row + 1] - b_ind; ++j) {
                    intType col = jb[j];
                    dataType val  = b[j];
                    std::cout << "B(" << row + b_ind << ", " << col << ") = " << val << std::endl;
                }
            }
        }

        good = true;

        q.wait_and_throw();
    }
    catch (sycl::exception const &e) {
        std::cout << "\t\tCaught synchronous SYCL exception:\n" << e.what() << std::endl;
        good = false;
    }
    catch (std::exception const &e) {
        std::cout << "\t\tCaught std exception:\n" << e.what() << std::endl;
        good = false;
    }

    q.wait();

    // backup cleaning of matrix handle and others for if exceptions happened
    if(csrA) oneapi::mkl::sparse::release_matrix_handle(q, &csrA);
    if(csrB) oneapi::mkl::sparse::release_matrix_handle(q, &csrB);

    q.wait();

    return good ? 0 : 1;
}

//
// Description of example setup, apis used and supported floating point type
// precisions
//
void print_example_banner()
{

    std::cout << "" << std::endl;
    std::cout << "###############################################################"
                 "#########"
              << std::endl;
    std::cout << "# Sparse Matrix Copy/Transpose/Conjugate-transpose Example: " << std::endl;
    std::cout << "# " << std::endl;
    std::cout << "#    B = op(A)" << std::endl;
    std::cout << "# " << std::endl;
    std::cout << "# where A and B are sparse matrices in CSR format" << std::endl;
    std::cout << "# " << std::endl;
    std::cout << "# Using apis:" << std::endl;
    std::cout << "#   sparse::omatcopy" << std::endl;
    std::cout << "# " << std::endl;
    std::cout << "#   sparse::init_matrix_handle" << std::endl;
    std::cout << "#   sparse::set_csr_data" << std::endl;
    std::cout << "#   sparse::release_matrix_handle" << std::endl;
    std::cout << "# " << std::endl;
    std::cout << "# Supported floating point type precisions:" << std::endl;
    std::cout << "#   float" << std::endl;
    std::cout << "#   double" << std::endl;
    std::cout << "#   std::complex<float>" << std::endl;
    std::cout << "#   std::complex<double>" << std::endl;
    std::cout << "# " << std::endl;
    std::cout << "###############################################################"
                 "#########"
              << std::endl;
    std::cout << std::endl;
}

//
// Main entry point for example.
//
// Dispatches to appropriate device types as set at build time with flag:
// -DSYCL_DEVICES_cpu -- only runs SYCL CPU implementation
// -DSYCL_DEVICES_gpu -- only runs SYCL GPU implementation
// -DSYCL_DEVICES_all (default) -- runs on all: cpu and gpu devices
//
//  For each device selected and each supported data type,
//  run_sparse_blas_example() is run with all supported data types,
//  if any fail, we move on to the next device.
//

int main(int argc, char **argv)
{
    print_example_banner();

    std::list<my_sycl_device_types> list_of_devices;
    set_list_of_devices(list_of_devices);

    int status = 0;
    for (auto it = list_of_devices.begin(); it != list_of_devices.end(); ++it) {
        try {
            sycl::device my_dev;
            bool my_dev_is_found = false;
            get_sycl_device(my_dev, my_dev_is_found, *it);

            if (my_dev_is_found) {
                std::cout << "Running tests on " << sycl_device_names[*it] << ".\n";

                // Catch asynchronous exceptions
                auto exception_handler = [](sycl::exception_list exceptions) {
                    for (std::exception_ptr const &e : exceptions) {
                        try {
                            std::rethrow_exception(e);
                        }
                        catch (sycl::exception const &e) {
                            std::cout << "Caught asynchronous SYCL exception: \n"
                                << e.what() << std::endl;
                        }
                    }
                };

                sycl::queue q(my_dev, exception_handler);

                std::cout << "\tRunning with single precision real data type:" << std::endl;
                status |= run_sparse_blas_example<float, std::int32_t>(q);

                if (my_dev.get_info<sycl::info::device::double_fp_config>().size() != 0) {
                    std::cout << "\tRunning with double precision real data type:" << std::endl;
                    status |= run_sparse_blas_example<double, std::int32_t>(q);
                }

                std::cout << "\tRunning with single precision complex data type:" << std::endl;
                status |= run_sparse_blas_example<std::complex<float>, std::int32_t>(q);

                if (my_dev.get_info<sycl::info::device::double_fp_config>().size() != 0) {
                    std::cout << "\tRunning with double precision complex data type:" << std::endl;
                    status |= run_sparse_blas_example<std::complex<double>, std::int32_t>(q);
                }

            }
            else {
#ifdef FAIL_ON_MISSING_DEVICES
                std::cout << "No " << sycl_device_names[*it]
                    << " devices found; Fail on missing devices "
                    "is enabled.\n";
                return 1;
#else
                std::cout << "No " << sycl_device_names[*it] << " devices found; skipping "
                    << sycl_device_names[*it] << " tests.\n";
#endif
            }
        }
        catch (sycl::exception const &e) {
            std::cout << "\t\tCaught SYCL exception at driver level: \n" << e.what() << std::endl;
            continue; // stop with device, but move on to other devices
        }
        catch (std::exception const &e) {
            std::cout << "\t\tCaught std exception at driver level: \n" << e.what() << std::endl;
            continue; // stop with device, but move on to other devices
        }

    } // for device

    mkl_free_buffers();
    return status;
}
